import pandas as pd
import torch
import typing as _typing
from . import (
    _abstract_views,
    _canonical_edge_type,
    _general_static_graph
)


class HeterogeneousNodesContainer:
    @property
    def node_types(self) -> _typing.AbstractSet[str]:
        raise NotImplementedError

    def remove_nodes(self, node_t: _typing.Optional[str]) -> 'HeterogeneousNodesContainer':
        raise NotImplementedError

    def reset_nodes(
            self, node_t: _typing.Optional[str],
            nodes_data: _typing.Mapping[str, torch.Tensor]
    ) -> 'HeterogeneousNodesContainer':
        raise NotImplementedError

    def set_data(
            self, node_t: _typing.Optional[str], data_key: str, data: torch.Tensor
    ) -> 'HeterogeneousNodesContainer':
        raise NotImplementedError

    def get_data(
            self, node_t: _typing.Optional[str] = ...,
            data_key: _typing.Optional[str] = ...
    ) -> _typing.Union[torch.Tensor, _typing.Mapping[str, torch.Tensor]]:
        raise NotImplementedError

    def delete_data(
            self, node_t: _typing.Optional[str], data_key: str
    ) -> 'HeterogeneousNodesContainer':
        raise TypeError

    def remove_data(
            self, node_t: _typing.Optional[str], data_key: str
    ) -> 'HeterogeneousNodesContainer':
        return self.delete_data(node_t, data_key)


class HeterogeneousNodesContainerImplementation(HeterogeneousNodesContainer):
    def __init__(self, data: _typing.Optional[_typing.Mapping[str, _typing.Mapping[str, torch.Tensor]]] = ...):
        self.__nodes_data: _typing.MutableMapping[str, _typing.MutableMapping[str, torch.Tensor]] = {}
        if data not in (None, Ellipsis) and isinstance(data, _typing.Mapping):
            for node_t, nodes_data in data.items():
                self.reset_nodes(node_t, nodes_data)

    @property
    def node_types(self) -> _typing.AbstractSet[str]:
        return self.__nodes_data.keys()

    def remove_nodes(self, node_t: _typing.Optional[str]) -> HeterogeneousNodesContainer:
        if not (node_t in (Ellipsis, None) or isinstance(node_t, str)):
            raise TypeError
        elif node_t in (Ellipsis, None):
            if len(self.node_types) == 0:
                return self
            elif len(self.node_types) == 1:
                del self.__nodes_data[tuple(self.node_types)[0]]
            else:
                _error_message: str = ' '.join((
                    "Unable to determine node type automatically,",
                    "possible cause is that the graph contains heterogeneous nodes,",
                    "node type must be specified for graph containing heterogeneous nodes."
                ))
                raise TypeError(_error_message)
        elif isinstance(node_t, str):
            try:
                del self.__nodes_data[node_t]
            except Exception:
                raise ValueError(f"nodes with type [{node_t}] NOT exists")
        return self

    def reset_nodes(
            self, node_t: _typing.Optional[str],
            nodes_data: _typing.Mapping[str, torch.Tensor]
    ) -> HeterogeneousNodesContainer:
        if not (node_t in (Ellipsis, None) or isinstance(node_t, str)):
            raise TypeError
        elif node_t in (Ellipsis, None) and len(self.node_types) > 1:
            _error_message: str = ' '.join((
                "Unable to determine node type automatically,",
                "possible cause is that the graph contains heterogeneous nodes,",
                "node type must be specified for graph containing heterogeneous nodes."
            ))
            raise TypeError(_error_message)
        elif isinstance(node_t, str) and ' ' in node_t:
            raise ValueError("node type must NOT contain space character (\' \').")
        __node_t: str = "" if node_t is Ellipsis else node_t

        num_nodes: int = ...
        for data_key, data_item in nodes_data.items():
            if not isinstance(data_key, str):
                raise TypeError
            if ' ' in data_key:
                raise ValueError("data key must NOT contain space character (\' \').")
            if not isinstance(data_item, torch.Tensor):
                raise TypeError
            if not data_item.dim() > 0:
                raise ValueError(
                    "data item MUST have at least one dimension, "
                    "and the first dimension corresponds to data for diverse nodes."
                )
            if not isinstance(num_nodes, int):
                num_nodes: int = data_item.size(0)
            if data_item.size(0) != num_nodes:
                raise ValueError
            self.__nodes_data[__node_t] = dict(nodes_data)
        return self

    def set_data(
            self, node_t: _typing.Optional[str], data_key: str, data: torch.Tensor
    ) -> HeterogeneousNodesContainer:
        if node_t in (Ellipsis, None):
            if len(self.node_types) == 0:
                __node_t: str = ""  # Default node type for homogeneous graph
            elif len(self.node_types) == 1:
                __node_t: str = list(self.node_types)[0]
            else:
                _error_message: str = ' '.join((
                    "Unable to determine node type automatically,",
                    "possible cause is that the graph contains heterogeneous nodes,",
                    "node type must be specified for graph containing heterogeneous nodes."
                ))
                raise TypeError(_error_message)
        elif isinstance(node_t, str):
            __node_t: str = node_t
        else:
            raise TypeError
        if not isinstance(data_key, str):
            raise TypeError
        if not isinstance(data, torch.Tensor):
            raise TypeError
        if ' ' in __node_t:
            raise ValueError
        if ' ' in data_key:
            raise ValueError
        if not data.dim() > 0:
            raise ValueError(
                "data item MUST have at least one dimension, "
                "and the first dimension corresponds to data for diverse nodes."
            )
        if __node_t not in self.node_types:
            self.__nodes_data[__node_t] = dict([(data_key, data)])
        else:
            obsolete_data: _typing.Optional[torch.Tensor] = self.__nodes_data[__node_t].get(data_key)
            if obsolete_data is not None and isinstance(obsolete_data, torch.Tensor):
                if data.size(0) != obsolete_data.size(0):
                    raise ValueError
            elif len(self.__nodes_data.get(__node_t)) > 0:
                num_nodes: int = self.__nodes_data[__node_t][list(self.__nodes_data[__node_t].keys())[0]].size(0)
                if data.size(0) != num_nodes:
                    raise ValueError
            self.__nodes_data[__node_t][data_key] = data
        return self

    def __get_data_for_specific_node_type(
            self, node_t: str, data_key: _typing.Optional[str] = ...
    ) -> _typing.Union[torch.Tensor, _typing.Mapping[str, torch.Tensor]]:
        if not isinstance(node_t, str):
            raise TypeError
        elif ' ' in node_t:
            raise ValueError
        if not (data_key in (Ellipsis, None) or isinstance(data_key, str)):
            raise TypeError
        elif isinstance(data_key, str) and ' ' in data_key:
            raise ValueError
        if node_t not in self.node_types:
            raise ValueError("Node type NOT exists")
        elif isinstance(data_key, str):
            data: _typing.Optional[torch.Tensor] = self.__nodes_data[node_t].get(data_key)
            if data is not None:
                return data
            else:
                raise KeyError(
                    f"Data with key [{data_key}] NOT exists "
                    f"for nodes with specific type [{node_t}]"
                )
        else:
            return self.__nodes_data[node_t]

    def __get_data_for_specific_data_key(
            self, data_key: str, node_t: _typing.Optional[str] = ...
    ) -> _typing.Union[torch.Tensor, _typing.Mapping[str, torch.Tensor]]:
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError
        if not (node_t in (Ellipsis, None) or isinstance(node_t, str)):
            raise TypeError
        elif isinstance(node_t, str) and ' ' in node_t:
            raise ValueError
        if isinstance(node_t, str):
            if node_t not in self.node_types:
                raise ValueError("Node type NOT exists")
            else:
                data: _typing.Optional[torch.Tensor] = (
                    self.__nodes_data[node_t].get(data_key)
                )
                if data is not None:
                    return data
                else:
                    raise KeyError(
                        f"Data with key [{data_key}] NOT exists "
                        f"for nodes with specific type [{node_t}]"
                    )
        else:
            if len(self.node_types) == 0:
                raise RuntimeError("Unable to get data from empty graph")
            elif len(self.node_types) == 1:
                __node_t: str = tuple(self.node_types)[0]
                __optional_data: _typing.Optional[torch.Tensor] = (
                    self.__nodes_data[__node_t].get(data_key)
                )
                if __optional_data is not None:
                    return __optional_data
                else:
                    raise KeyError(f"Data with key [{data_key}] NOT exists")
            else:
                __result: _typing.Dict[str, torch.Tensor] = {}
                for __node_t, __nodes_data in self.__nodes_data.items():
                    __optional_data: _typing.Optional[torch.Tensor] = (
                        __nodes_data.get(data_key)
                    )
                    if (
                            __optional_data is not None and
                            isinstance(__optional_data, torch.Tensor)
                    ):
                        __result[__node_t] = __optional_data
                if len(__result):
                    return __result
                else:
                    raise KeyError(f"Data with key [{data_key}] NOT exists")

    def get_data(
            self, node_t: _typing.Optional[str] = ...,
            data_key: _typing.Optional[str] = ...
    ) -> _typing.Union[torch.Tensor, _typing.Mapping[str, torch.Tensor]]:
        if not (node_t in (Ellipsis, None) or isinstance(node_t, str)):
            raise TypeError
        elif isinstance(node_t, str) and ' ' in node_t:
            raise ValueError
        if not (data_key in (Ellipsis, None) or isinstance(data_key, str)):
            raise TypeError
        elif isinstance(data_key, str) and ' ' in data_key:
            raise ValueError
        if isinstance(node_t, str):
            return self.__get_data_for_specific_node_type(node_t, data_key)
        elif node_t in (Ellipsis, None) and isinstance(data_key, str):
            return self.__get_data_for_specific_data_key(data_key)
        elif node_t in (Ellipsis, None) and data_key in (Ellipsis, None):
            if len(self.node_types) == 1:
                __node_t: str = tuple(self.node_types)[0]
                return self.__get_data_for_specific_node_type(__node_t)
            else:
                raise TypeError(
                    "Unable to determine node type automatically, "
                    "possible cause is that the graph contains heterogeneous nodes or is empty, "
                    "node type must be specified for graph containing heterogeneous nodes."
                )

    def delete_data(
            self, node_t: _typing.Optional[str], data_key: str
    ) -> HeterogeneousNodesContainer:
        if not (node_t in (Ellipsis, None) or isinstance(node_t, str)):
            raise TypeError
        elif node_t in (Ellipsis, None):
            if len(self.node_types) == 1:
                __node_t: str = tuple(self.node_types)[0]
            else:
                raise TypeError(
                    "Unable to determine node type automatically, "
                    "possible cause is that the graph contains heterogeneous nodes or is empty, "
                    "node type must be specified for graph containing heterogeneous nodes."
                )
        elif isinstance(node_t, str):
            if node_t in self.node_types:
                __node_t: str = node_t
            else:
                raise ValueError("node type NOT exists")
        else:
            raise TypeError
        if not isinstance(data_key, str):
            raise TypeError
        elif data_key not in self.__nodes_data.get(__node_t):
            raise KeyError(
                f"Data with key [{data_key}] NOT exists for nodes with type [{__node_t}]"
            )
        else:
            self.__nodes_data[__node_t].__delitem__(data_key)
            if len(self.__nodes_data.get(__node_t)) == 0:
                del self.__nodes_data[__node_t]
        return self


class _SpecificTypedNodeDataView(_abstract_views.SpecificTypedNodeDataView):
    def __init__(
            self, heterogeneous_nodes_container: HeterogeneousNodesContainer,
            node_type: _typing.Optional[str]
    ):
        if not isinstance(heterogeneous_nodes_container, HeterogeneousNodesContainer):
            raise TypeError
        else:
            self._heterogeneous_nodes_container: HeterogeneousNodesContainer = (
                heterogeneous_nodes_container
            )
        if not (isinstance(node_type, str) or node_type in (Ellipsis, None)):
            raise TypeError
        elif isinstance(node_type, str):
            if node_type not in self._heterogeneous_nodes_container.node_types:
                raise ValueError("Invalid node type")
        self.__node_t: _typing.Optional[str] = node_type

    def __getitem__(self, data_key: str) -> torch.Tensor:
        return self._heterogeneous_nodes_container.get_data(self.__node_t, data_key)

    def __setitem__(self, data_key: str, value: torch.Tensor):
        self._heterogeneous_nodes_container.set_data(self.__node_t, data_key, value)

    def __delitem__(self, data_key: str) -> None:
        self._heterogeneous_nodes_container.delete_data(self.__node_t, data_key)

    def __len__(self) -> int:
        return len(self._heterogeneous_nodes_container.get_data(self.__node_t))

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self._heterogeneous_nodes_container.get_data(self.__node_t))


class _SpecificTypedNodeView(_abstract_views.SpecificTypedNodeView):
    def __init__(
            self, nodes_container: HeterogeneousNodesContainer,
            node_t: _typing.Optional[str]
    ):
        self._heterogeneous_nodes_container: HeterogeneousNodesContainer = nodes_container
        self.__node_t: _typing.Optional[str] = node_t

    @property
    def data(self) -> _SpecificTypedNodeDataView:
        return _SpecificTypedNodeDataView(self._heterogeneous_nodes_container, self.__node_t)

    @data.setter
    def data(self, nodes_data: _typing.Mapping[str, torch.Tensor]):
        self._heterogeneous_nodes_container.reset_nodes(self.__node_t, nodes_data)


class _HeterogeneousNodeView(_abstract_views.HeterogeneousNodeView):
    def __init__(self, nodes_container: HeterogeneousNodesContainer):
        self._heterogeneous_nodes_container: HeterogeneousNodesContainer = nodes_container

    def __getitem__(self, node_type: _typing.Optional[str]) -> _SpecificTypedNodeView:
        return _SpecificTypedNodeView(self._heterogeneous_nodes_container, node_type)

    def __setitem__(
            self, node_t: _typing.Optional[str],
            nodes_data: _typing.Mapping[str, torch.Tensor]
    ) -> None:
        self._heterogeneous_nodes_container.reset_nodes(node_t, nodes_data)

    def __delitem__(self, node_t: _typing.Optional[str]):
        self._heterogeneous_nodes_container.remove_nodes(node_t)

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self._heterogeneous_nodes_container.node_types)

    @property
    def data(self) -> _SpecificTypedNodeDataView:
        return _SpecificTypedNodeDataView(self._heterogeneous_nodes_container, ...)

    @data.setter
    def data(self, nodes_data: _typing.Mapping[str, torch.Tensor]):
        self._heterogeneous_nodes_container.reset_nodes(..., nodes_data)

    @property
    def is_homogeneous(self) -> bool:
        return len(self._heterogeneous_nodes_container.node_types) <= 1


class HomogeneousEdgesContainer:
    @property
    def connections(self) -> torch.Tensor:
        raise NotImplementedError

    @property
    def data_keys(self) -> _typing.Iterable[str]:
        raise NotImplementedError

    def get_data(
            self, data_key: _typing.Optional[str] = ...
    ) -> _typing.Union[torch.Tensor, _typing.Mapping[str, torch.Tensor]]:
        raise NotImplementedError

    def set_data(self, data_key: str, data: torch.Tensor):
        raise NotImplementedError

    def delete_data(self, data_key: str):
        raise NotImplementedError


class HomogeneousEdgesContainerImplementation(HomogeneousEdgesContainer):
    def __init__(
            self, edge_connections: torch.Tensor,
            data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        if not isinstance(edge_connections, torch.Tensor):
            raise TypeError
        if not (data in (Ellipsis, None) or isinstance(data, _typing.Mapping)):
            raise TypeError
        if not (
                edge_connections.dtype == torch.int64 and
                edge_connections.dim() == edge_connections.size(0) == 2
        ):
            raise ValueError
        self.__connections: torch.Tensor = edge_connections
        if not isinstance(data, _typing.Mapping):
            self.__data: _typing.MutableMapping[str, torch.Tensor] = {}
        else:
            for data_key, data_item in data.items():
                if not isinstance(data_key, str):
                    raise TypeError
                if not isinstance(data_item, torch.Tensor):
                    raise TypeError
                if ' ' in data_key:
                    raise ValueError
                if not data_item.dim() > 0:
                    raise ValueError
                if data_item.size(0) != self.__connections.size(1):
                    raise ValueError
            self.__data: _typing.MutableMapping[str, torch.Tensor] = dict(data)

    @property
    def connections(self) -> torch.Tensor:
        return self.__connections

    @property
    def data_keys(self) -> _typing.Iterable[str]:
        return self.__data.keys()

    def set_data(self, data_key: str, data: torch.Tensor) -> HomogeneousEdgesContainer:
        if not isinstance(data_key, str):
            raise TypeError
        if not isinstance(data, torch.Tensor):
            raise TypeError
        if ' ' in data_key:
            raise ValueError
        if data.dim() == 0 or data.size(0) != self.__connections.size(1):
            raise ValueError
        self.__data[data_key] = data
        return self

    def get_data(
            self, data_key: _typing.Optional[str] = ...
    ) -> _typing.Union[torch.Tensor, _typing.Mapping[str, torch.Tensor]]:
        if not (data_key in (Ellipsis, None) or isinstance(data_key, str)):
            raise TypeError
        if isinstance(data_key, str):
            if ' ' in data_key:
                raise ValueError
            temp: _typing.Optional[torch.Tensor] = self.__data.get(data_key)
            if temp is None:
                raise KeyError(f"Data with key [{data_key}] NOT exists")
            else:
                return temp
        else:
            return dict(self.__data)

    def delete_data(self, data_key: str) -> HomogeneousEdgesContainer:
        if not isinstance(data_key, str):
            raise TypeError
        if ' ' in data_key:
            raise ValueError
        try:
            del self.__data[data_key]
        finally:
            return self


class HeterogeneousEdgesAggregation(
    _typing.MutableMapping[
        _typing.Union[str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType],
        HomogeneousEdgesContainer
    ]
):
    def __setitem__(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType],
            edges: _typing.Union[HomogeneousEdgesContainer, torch.LongTensor]
    ) -> None:
        self._set_edges(edge_t, edges)

    def __delitem__(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType]
    ) -> None:
        self._delete_edges(edge_t)

    def __getitem__(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType] = ...
    ) -> HomogeneousEdgesContainer:
        return self._get_edges(edge_t)

    def __len__(self) -> int:
        return len(list(self._edge_types))

    def __iter__(self) -> _typing.Iterator[_canonical_edge_type.CanonicalEdgeType]:
        return iter(self._edge_types)

    @property
    def _edge_types(self) -> _typing.Iterable[_canonical_edge_type.CanonicalEdgeType]:
        raise NotImplementedError

    def _get_edges(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType] = ...
    ) -> HomogeneousEdgesContainer:
        raise NotImplementedError

    def _set_edges(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType],
            edges: _typing.Union[HomogeneousEdgesContainer, torch.LongTensor]
    ):
        raise NotImplementedError

    def _delete_edges(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType]
    ) -> None:
        raise NotImplementedError


class HeterogeneousEdgesAggregationImplementation(HeterogeneousEdgesAggregation):
    def __init__(self):
        self.__heterogeneous_edges_data_frame: pd.DataFrame = pd.DataFrame(
            columns=('s', 'r', 't', 'edges'),
        )

    @property
    def _edge_types(self) -> _typing.Iterable[_canonical_edge_type.CanonicalEdgeType]:
        return [
            _canonical_edge_type.CanonicalEdgeType(getattr(row_tuple, 's'), getattr(row_tuple, 'r'), getattr(row_tuple, 't'))
            for row_tuple in self.__heterogeneous_edges_data_frame.itertuples(False, name="Edge")
        ]

    def _get_edges(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType] = ...
    ) -> HomogeneousEdgesContainer:
        if edge_t in (Ellipsis, None):
            if len(self.__heterogeneous_edges_data_frame) == 0:
                raise ValueError("The graph contains no edges")
            elif len(self.__heterogeneous_edges_data_frame) == 1:
                return self.__heterogeneous_edges_data_frame.iloc[0]['edges']
            else:
                raise ValueError(
                    "Unable to automatically determine edge type "
                    "since the graph contains multiple edge types"
                )
        elif isinstance(edge_t, str):
            if ' ' in edge_t:
                raise ValueError
            if len(
                    self.__heterogeneous_edges_data_frame.loc[
                        self.__heterogeneous_edges_data_frame['r'] == edge_t
                    ]
            ) == 0:
                raise ValueError(f"The graph has NOT edge with relation type as {edge_t}")
            elif len(
                    self.__heterogeneous_edges_data_frame.loc[
                        self.__heterogeneous_edges_data_frame['r'] == edge_t
                    ]
            ) == 1:
                temp: HomogeneousEdgesContainer = self.__heterogeneous_edges_data_frame.loc[
                    self.__heterogeneous_edges_data_frame['r'] == edge_t, 'edges'
                ]
                if not isinstance(temp, HomogeneousEdgesContainer):
                    raise RuntimeError
                else:
                    return temp
            else:
                raise ValueError(
                    f"Unable to determine canonical edge type by relation type \"{edge_t}\", "
                    f"since the graph contains multiple edge types with relation type as \"{edge_t}\""
                )
        elif isinstance(edge_t, _typing.Tuple) or isinstance(edge_t, _canonical_edge_type.CanonicalEdgeType):
            if isinstance(edge_t, _typing.Tuple) and not (
                    len(edge_t) == 3 and
                    isinstance(edge_t[0], str) and
                    isinstance(edge_t[1], str) and
                    isinstance(edge_t[2], str) and
                    ' ' not in edge_t[0] and ' ' not in edge_t[1] and ' ' not in edge_t[2]
            ):
                raise TypeError("Illegal canonical edge type")
            __edge_t: _typing.Tuple[str, str, str] = (
                (edge_t.source_node_type, edge_t.relation_type, edge_t.target_node_type)
                if isinstance(edge_t, _canonical_edge_type.CanonicalEdgeType) else edge_t
            )
            partial_data_frame: pd.DataFrame = self.__heterogeneous_edges_data_frame.loc[
                (self.__heterogeneous_edges_data_frame['s'] == __edge_t[0]) &
                (self.__heterogeneous_edges_data_frame['r'] == __edge_t[1]) &
                (self.__heterogeneous_edges_data_frame['t'] == __edge_t[2])
                ]
            if len(partial_data_frame) == 0:
                raise ValueError
            elif len(partial_data_frame) == 1:
                temp: HomogeneousEdgesContainer = partial_data_frame.iloc[0]['edges']
                if not isinstance(temp, HomogeneousEdgesContainer):
                    raise RuntimeError
                else:
                    return temp
            else:
                raise RuntimeError

    def _set_edges(
            self,
            edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType],
            edges: _typing.Union[HomogeneousEdgesContainer, torch.LongTensor]
    ):
        if not (isinstance(edges, HomogeneousEdgesContainer) or isinstance(edges, torch.Tensor)):
            raise TypeError
        if edge_t in (Ellipsis, None):
            if len(self.__heterogeneous_edges_data_frame) == 0:
                self.__heterogeneous_edges_data_frame: pd.DataFrame = (
                    self.__heterogeneous_edges_data_frame.append(
                        pd.DataFrame(
                            {
                                's': [''], 'r': [''], 't': [''],
                                'edges': [
                                    edges if isinstance(edges, HomogeneousEdgesContainer)
                                    else HomogeneousEdgesContainerImplementation(edges)
                                ]
                            }
                        )
                    )
                )
            elif len(self.__heterogeneous_edges_data_frame) == 1:
                self.__heterogeneous_edges_data_frame.iloc[0]['edges'] = (
                    edges if isinstance(edges, HomogeneousEdgesContainer)
                    else HomogeneousEdgesContainerImplementation(edges)
                )
            else:
                raise ValueError(
                    "Unable to set edges for heterogeneous graph consist of multiple edge types "
                    "with automatically determined edge type"
                )
        elif isinstance(edge_t, str):
            if ' ' in edge_t:
                raise ValueError
            if len(
                    self.__heterogeneous_edges_data_frame.loc[
                        self.__heterogeneous_edges_data_frame['r'] == edge_t
                    ]
            ) == 1:
                self.__heterogeneous_edges_data_frame.loc[
                    self.__heterogeneous_edges_data_frame['r'] == edge_t, 'edges'
                ] = (
                    edges if isinstance(edges, HomogeneousEdgesContainer)
                    else HomogeneousEdgesContainerImplementation(edges)
                )
            else:
                raise RuntimeError
        elif isinstance(edge_t, _typing.Tuple) or isinstance(edge_t, _canonical_edge_type.CanonicalEdgeType):
            if isinstance(edge_t, _typing.Tuple) and not (
                    len(edge_t) == 3 and
                    isinstance(edge_t[0], str) and
                    isinstance(edge_t[1], str) and
                    isinstance(edge_t[2], str) and
                    ' ' not in edge_t[0] and ' ' not in edge_t[1] and ' ' not in edge_t[2]
            ):
                raise TypeError("Illegal canonical edge type")
            __edge_t: _typing.Tuple[str, str, str] = (
                (edge_t.source_node_type, edge_t.relation_type, edge_t.target_node_type)
                if isinstance(edge_t, _canonical_edge_type.CanonicalEdgeType) else edge_t
            )
            if len(
                    self.__heterogeneous_edges_data_frame.loc[
                        (self.__heterogeneous_edges_data_frame['s'] == __edge_t[0]) &
                        (self.__heterogeneous_edges_data_frame['r'] == __edge_t[1]) &
                        (self.__heterogeneous_edges_data_frame['t'] == __edge_t[2])
                    ]
            ) == 0:
                self.__heterogeneous_edges_data_frame: pd.DataFrame = (
                    self.__heterogeneous_edges_data_frame.append(
                        pd.DataFrame(
                            {
                                's': [__edge_t[0]],
                                'r': [__edge_t[1]],
                                't': [__edge_t[2]],
                                'edges': [
                                    edges if isinstance(edges, HomogeneousEdgesContainer)
                                    else HomogeneousEdgesContainerImplementation(edges)
                                ]
                            }
                        )
                    )
                )
            elif len(
                    self.__heterogeneous_edges_data_frame.loc[
                        (self.__heterogeneous_edges_data_frame['s'] == __edge_t[0]) &
                        (self.__heterogeneous_edges_data_frame['r'] == __edge_t[1]) &
                        (self.__heterogeneous_edges_data_frame['t'] == __edge_t[2])
                    ]
            ) == 1:
                self.__heterogeneous_edges_data_frame.loc[
                    (self.__heterogeneous_edges_data_frame['s'] == __edge_t[0]) &
                    (self.__heterogeneous_edges_data_frame['r'] == __edge_t[1]) &
                    (self.__heterogeneous_edges_data_frame['t'] == __edge_t[2]),
                    'edges'
                ] = (
                    edges if isinstance(edges, HomogeneousEdgesContainer)
                    else HomogeneousEdgesContainerImplementation(edges)
                )
            else:
                raise RuntimeError  # todo: Unable to determine error
        else:
            raise TypeError("Unsupported edge type")

    def _delete_edges(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType] = ...
    ) -> None:
        if edge_t in (Ellipsis, None):
            if len(self.__heterogeneous_edges_data_frame) == 1:
                self.__heterogeneous_edges_data_frame.drop(
                    self.__heterogeneous_edges_data_frame.index[0], inplace=True
                )
            elif len(self.__heterogeneous_edges_data_frame) > 1:
                raise ValueError("Edge Type must be specified for graph containing heterogeneous edges")
        elif isinstance(edge_t, str):
            if ' ' in edge_t:
                raise ValueError
            if len(self.__heterogeneous_edges_data_frame) > 0:
                self.__heterogeneous_edges_data_frame: pd.DataFrame = (
                    self.__heterogeneous_edges_data_frame[
                        self.__heterogeneous_edges_data_frame['r'] != edge_t
                        ].reset_index(drop=True)
                )
        elif isinstance(edge_t, _typing.Tuple) or isinstance(edge_t, _canonical_edge_type.CanonicalEdgeType):
            if isinstance(edge_t, _typing.Tuple) and not (
                    len(edge_t) == 3 and
                    isinstance(edge_t[0], str) and
                    isinstance(edge_t[1], str) and
                    isinstance(edge_t[2], str) and
                    ' ' not in edge_t[0] and ' ' not in edge_t[1] and ' ' not in edge_t[2]
            ):
                raise TypeError("Illegal canonical edge type")
            __edge_t: _typing.Tuple[str, str, str] = (
                (edge_t.source_node_type, edge_t.relation_type, edge_t.target_node_type)
                if isinstance(edge_t, _canonical_edge_type.CanonicalEdgeType) else edge_t
            )
            if len(self.__heterogeneous_edges_data_frame) > 0:
                self.__heterogeneous_edges_data_frame: pd.DataFrame = (
                    self.__heterogeneous_edges_data_frame[
                        (self.__heterogeneous_edges_data_frame['s'] != edge_t) |
                        (self.__heterogeneous_edges_data_frame['r'] != edge_t) |
                        (self.__heterogeneous_edges_data_frame['t'] != edge_t)
                        ].reset_index(drop=True)
                )
        else:
            raise TypeError("Unsupported edge type")


class _HomogeneousEdgesDataView(_abstract_views.HomogeneousEdgesDataView):
    def __init__(self, homogeneous_edges_container: HomogeneousEdgesContainer):
        if not isinstance(homogeneous_edges_container, HomogeneousEdgesContainer):
            raise TypeError
        self._homogeneous_edges_container: HomogeneousEdgesContainer = homogeneous_edges_container

    def __getitem__(self, data_key: str) -> torch.Tensor:
        if not isinstance(data_key, str):
            raise TypeError
        if ' ' in data_key:
            raise ValueError
        return self._homogeneous_edges_container.get_data(data_key)

    def __setitem__(self, data_key: str, data: torch.Tensor):
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError
        if not isinstance(data, torch.Tensor):
            raise TypeError
        elif not data.dim() > 0:
            raise ValueError
        self._homogeneous_edges_container.set_data(data_key, data)

    def __delitem__(self, data_key: str):
        if not isinstance(data_key, str):
            raise TypeError
        elif ' ' in data_key:
            raise ValueError
        self._homogeneous_edges_container.delete_data(data_key)

    def __len__(self):
        return len(list(self._homogeneous_edges_container.data_keys))

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self._homogeneous_edges_container.data_keys)


class _SpecificTypedHomogeneousEdgesView(_abstract_views.HomogeneousEdgesView):
    def __init__(self, homogeneous_edges_container: HomogeneousEdgesContainer):
        if not isinstance(homogeneous_edges_container, HomogeneousEdgesContainer):
            raise TypeError
        self._homogeneous_edges_container: HomogeneousEdgesContainer = homogeneous_edges_container

    @property
    def connections(self) -> torch.Tensor:
        return self._homogeneous_edges_container.connections

    @property
    def data(self) -> _HomogeneousEdgesDataView:
        return _HomogeneousEdgesDataView(self._homogeneous_edges_container)


class _HeterogeneousEdgesView(_abstract_views.HeterogeneousEdgesView):
    def __init__(self, _heterogeneous_edges_aggregation: HeterogeneousEdgesAggregation):
        if not isinstance(_heterogeneous_edges_aggregation, HeterogeneousEdgesAggregation):
            raise TypeError
        self._heterogeneous_edges_aggregation: HeterogeneousEdgesAggregation = (
            _heterogeneous_edges_aggregation
        )

    def __getitem__(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType]
    ) -> _SpecificTypedHomogeneousEdgesView:
        return _SpecificTypedHomogeneousEdgesView(self._heterogeneous_edges_aggregation[edge_t])

    def __setitem__(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType],
            edges: _typing.Union[HomogeneousEdgesContainer, torch.LongTensor]
    ):
        self._heterogeneous_edges_aggregation[edge_t] = edges

    def __delitem__(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType]
    ):
        del self._heterogeneous_edges_aggregation[edge_t]

    def __len__(self) -> int:
        return len(self._heterogeneous_edges_aggregation)

    def __iter__(self) -> _typing.Iterator[_canonical_edge_type.CanonicalEdgeType]:
        return iter(self._heterogeneous_edges_aggregation)

    def __contains__(self, edge_type: _typing.Union[str, _typing.Tuple[str, str, str], _canonical_edge_type.CanonicalEdgeType]) -> bool:
        if isinstance(edge_type, str):
            if ' ' in edge_type:
                raise ValueError
            else:
                for existing_edge_type in self:
                    if existing_edge_type.relation_type == edge_type:
                        return True
                return False
        elif isinstance(edge_type, _typing.Tuple):
            if not (
                    len(edge_type) == 3 and
                    all([(isinstance(t, str) and ' ' not in t) for t in edge_type])
            ):
                raise TypeError
            else:
                for existing_edge_type in self:
                    if existing_edge_type.__eq__(edge_type):
                        return True
                return False
        elif isinstance(edge_type, _canonical_edge_type.CanonicalEdgeType):
            for existing_edge_type in self:
                if existing_edge_type == edge_type:
                    return True
            return False
        else:
            raise TypeError

    @property
    def connections(self) -> torch.Tensor:
        return self[...].connections

    @property
    def data(self) -> _HomogeneousEdgesDataView:
        return self[...].data

    @property
    def is_homogeneous(self) -> bool:
        return len(self) <= 1

    def set(
            self, edge_t: _typing.Union[None, str, _typing.Tuple[str, str, str]],
            connections: torch.LongTensor, data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        self[edge_t] = HomogeneousEdgesContainerImplementation(connections, data)


class _StaticGraphDataContainer(_typing.MutableMapping[str, torch.Tensor]):
    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        raise NotImplementedError

    def __delitem__(self, data_key: str) -> None:
        raise NotImplementedError

    def __getitem__(self, data_key: str) -> torch.Tensor:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def __iter__(self) -> _typing.Iterator[str]:
        raise NotImplementedError


class StaticGraphDataAggregation(_StaticGraphDataContainer):
    def __init__(
            self, graph_data: _typing.Optional[_typing.Mapping[str, torch.Tensor]] = ...
    ):
        self.__data: _typing.MutableMapping[str, torch.Tensor] = (
            dict(graph_data) if isinstance(graph_data, _typing.Mapping)
            else {}
        )

    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        self.__data[data_key] = data

    def __delitem__(self, data_key: str) -> None:
        del self.__data[data_key]

    def __getitem__(self, data_key: str) -> torch.Tensor:
        return self.__data[data_key]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self.__data)


class _StaticGraphDataView(_abstract_views.GraphDataView):
    def __init__(self, graph_data_container: _StaticGraphDataContainer):
        self.__graph_data_container: _StaticGraphDataContainer = (
            graph_data_container
        )

    def __setitem__(self, data_key: str, data: torch.Tensor) -> None:
        self.__graph_data_container[data_key] = data

    def __delitem__(self, data_key: str) -> None:
        del self.__graph_data_container[data_key]

    def __getitem__(self, data_key: str) -> torch.Tensor:
        return self.__graph_data_container[data_key]

    def __len__(self) -> int:
        return len(self.__graph_data_container)

    def __iter__(self) -> _typing.Iterator[str]:
        return iter(self.__graph_data_container)


class GeneralStaticGraphImplementation(_general_static_graph.GeneralStaticGraph):
    def __init__(
            self, _heterogeneous_nodes_container: _typing.Optional[HeterogeneousNodesContainer] = ...,
            _heterogeneous_edges_aggregation: _typing.Optional[HeterogeneousEdgesAggregation] = ...,
            graph_data_container: _typing.Optional[_StaticGraphDataContainer] = ...
    ):
        self._static_graph_data_container: _StaticGraphDataContainer = (
            graph_data_container
            if isinstance(graph_data_container, _StaticGraphDataContainer)
            else StaticGraphDataAggregation()
        )
        self._heterogeneous_nodes_container: HeterogeneousNodesContainer = (
            _heterogeneous_nodes_container
            if isinstance(_heterogeneous_nodes_container, HeterogeneousNodesContainer)
            else HeterogeneousNodesContainerImplementation()
        )
        self._heterogeneous_edges_aggregation: HeterogeneousEdgesAggregation = (
            _heterogeneous_edges_aggregation
            if isinstance(_heterogeneous_edges_aggregation, HeterogeneousEdgesAggregation)
            else HeterogeneousEdgesAggregationImplementation()
        )

    @property
    def nodes(self) -> _HeterogeneousNodeView:
        return _HeterogeneousNodeView(self._heterogeneous_nodes_container)

    @property
    def edges(self) -> _HeterogeneousEdgesView:
        return _HeterogeneousEdgesView(self._heterogeneous_edges_aggregation)

    @property
    def data(self) -> _StaticGraphDataView:
        return _StaticGraphDataView(self._static_graph_data_container)



# "Graph differentiable architecture search with structure optimization" NeurIPS 21'

import logging

import torch
import torch.optim
import torch.nn as nn
import torch.nn.functional as F

from . import register_nas_algo
from .base import BaseNAS
from ..estimator.base import BaseEstimator
from ..space import BaseSpace
from ..utils import replace_layer_choice, replace_input_choice
from ...model.base import BaseAutoModel

from torch.autograd import Variable
import numpy as np
import time
import copy
import torch.optim as optim
import scipy.sparse as sp

_logger = logging.getLogger(__name__)

@register_nas_algo("gasso")
class Gasso(BaseNAS):
    """
    GASSO trainer.

    Parameters
    ----------
    num_epochs : int
        Number of epochs planned for training.
    warmup_epochs : int
        Number of epochs planned for warming up.
    workers : int
        Workers for data loading.
    model_lr : float
        Learning rate to optimize the model.
    model_wd : float
        Weight decay to optimize the model.
    arch_lr : float
        Learning rate to optimize the architecture.
    stru_lr : float
        Learning rate to optimize the structure.
    lamb : float
        The parameter to control the influence of hidden feature smoothness
    device : str or torch.device
        The device of the whole process
    """
    def __init__(
        self,
        num_epochs=250,
        warmup_epochs=10,
        model_lr=0.01,
        model_wd=1e-4,
        arch_lr = 0.03,
        stru_lr = 0.04,
        lamb = 0.6,
        device="auto",
    ):
        super().__init__(device=device)
        self.num_epochs = num_epochs
        self.warmup_epochs = warmup_epochs
        self.model_lr = model_lr
        self.model_wd = model_wd
        self.arch_lr = arch_lr
        self.stru_lr = stru_lr
        self.lamb = lamb

    def train_stru(self, model, optimizer, data):
        # forward
        model.train()
        data[0].adj = self.adjs
        logits = model(data[0]).detach()
        loss = 0
        for adj in self.adjs:
            e1 = adj[0][0]
            e2 = adj[0][1]
            ew = adj[1]
            diff = (logits[e1] - logits[e2]).pow(2).sum(1)
            smooth = (diff * torch.sigmoid(ew)).sum()
            dist = (ew * ew).sum()
            loss += self.lamb * smooth + dist
            
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        train_loss = loss.item()
        del logits

    def _infer(self, model: BaseSpace, dataset, estimator: BaseEstimator, mask="train"):
        dataset[0].adj = self.adjs
        metric, loss = estimator.infer(model, dataset, mask=mask)
        return metric, loss

    def prepare(self, dset):
        """Train Pro-GNN.
        """
        data = dset[0]
        self.ews = []
        self.edges = data.edge_index.to(self.device)
        edge_weight = torch.ones(self.edges.size(1)).to(self.device)

        self.adjs = []
        for i in range(self.steps):
            edge_weight = Variable(edge_weight * 1.0, requires_grad = True).to(self.device)
            self.ews.append(edge_weight)
            self.adjs.append((self.edges, edge_weight))

    def fit(self, data):
        self.optimizer = optim.Adam(self.space.parameters(), lr=self.model_lr, weight_decay=self.model_wd)
        self.arch_optimizer = optim.Adam(self.space.arch_parameters(),
                                              lr=self.arch_lr, betas=(0.5, 0.999))
        self.stru_optimizer = optim.SGD(self.ews, lr=self.stru_lr)

        # Train model
        best_performance = 0
        min_val_loss = float("inf")
        min_train_loss = float("inf")

        t_total = time.time()
        for epoch in range(self.num_epochs):
            self.space.train()
            self.optimizer.zero_grad()
            _, loss = self._infer(self.space, data, self.estimator, "train")
            loss.backward()
            self.optimizer.step()

            if epoch <20:
                continue
            self.train_stru(self.space, self.stru_optimizer, data)
            
            self.arch_optimizer.zero_grad()
            _, loss = self._infer(self.space, data, self.estimator, "train")
            loss.backward()
            self.arch_optimizer.step()

            self.space.eval()
            train_acc, _ = self._infer(self.space, data, self.estimator, "train")
            val_acc, val_loss = self._infer(self.space, data, self.estimator, "val")
            if val_loss < min_val_loss:
                min_val_loss = val_loss
                best_performance = val_acc
                self.space.keep_prediction()
            #print("acc:" + str(train_acc) + " val_acc" + str(val_acc))

        return best_performance, min_val_loss

    def search(self, space: BaseSpace, dataset, estimator):
        self.estimator = estimator
        self.space = space.to(self.device)
        self.steps = space.steps
        self.prepare(dataset)
        perf, val_loss = self.fit(dataset)
        return space.parse_model(None, self.device)


from .base import BaseGenerator
from .graphlet import GeGraphlet
from .eigen import GeEigen
from .page_rank import GePageRank
from .pyg import (
    register_pyg,
    PYGGenerator,
    pygfunc,
    PYGLocalDegreeProfile,
    PYGNormalizeFeatures,
    PYGOneHotDegree,
)

__all__ = [
    "BaseGenerator",
    "GeGraphlet",
    "GeEigen",
    "GePageRank",
    "register_pyg",
    "pygfunc",
    "PYGGenerator",
    "PYGLocalDegreeProfile",
    "PYGNormalizeFeatures",
    "PYGOneHotDegree",
]



from .base import BaseFeature
from .base import BaseFeatureEngineer

FEATURE_DICT = {}


def register_feature(name):
    def register_feature_cls(cls):
        if name in FEATURE_DICT:
            raise ValueError(
                "Cannot register duplicate feature engineer ({})".format(name)
            )
        # if not issubclass(cls, BaseFeatureEngineer):
        if not issubclass(cls, BaseFeature):
            raise ValueError(
                "Trainer ({}: {}) must extend BaseFeatureEngineer".format(
                    name, cls.__name__
                )
            )
        FEATURE_DICT[name] = cls
        return cls

    return register_feature_cls


from .auto_feature import AutoFeatureEngineer

from .generators import (
    BaseGenerator,
    GeGraphlet,
    GeEigen,
    GePageRank,
    register_pyg,
    pygfunc,
    PYGGenerator,
    PYGLocalDegreeProfile,
    PYGNormalizeFeatures,
    PYGOneHotDegree,
)

from .selectors import BaseSelector, SeFilterConstant, SeGBDT

from .graph import (
    BaseGraph,
    SgNetLSD,
    register_nx,
    NxGraph,
    nxfunc,
    NxLargeCliqueSize,
    NxAverageClusteringApproximate,
    NxDegreeAssortativityCoefficient,
    NxDegreePearsonCorrelationCoefficient,
    NxHasBridge,
    NxGraphCliqueNumber,
    NxGraphNumberOfCliques,
    NxTransitivity,
    NxAverageClustering,
    NxIsConnected,
    NxNumberConnectedComponents,
    NxIsDistanceRegular,
    NxLocalEfficiency,
    NxGlobalEfficiency,
    NxIsEulerian,
)

__all__ = [
    "BaseFeatureEngineer",
    "AutoFeatureEngineer",
    "BaseFeature",
    "BaseGenerator",
    "GeGraphlet",
    "GeEigen",
    "GePageRank",
    "register_pyg",
    "pygfunc",
    "PYGGenerator",
    "PYGLocalDegreeProfile",
    "PYGNormalizeFeatures",
    "PYGOneHotDegree",
    "BaseSelector",
    "SeFilterConstant",
    "SeGBDT",
    "BaseGraph",
    "SgNetLSD",
    "register_nx",
    "NxGraph",
    "nxfunc",
    "NxLargeCliqueSize",
    "NxAverageClusteringApproximate",
    "NxDegreeAssortativityCoefficient",
    "NxDegreePearsonCorrelationCoefficient",
    "NxHasBridge",
    "NxGraphCliqueNumber",
    "NxGraphNumberOfCliques",
    "NxTransitivity",
    "NxAverageClustering",
    "NxIsConnected",
    "NxNumberConnectedComponents",
    "NxIsDistanceRegular",
    "NxLocalEfficiency",
    "NxGlobalEfficiency",
    "NxIsEulerian",
]



from .. import _data_preprocessor
class StructureEngineer(_data_preprocessor.DataPreprocessor):
    ...

import torch
from ....utils import get_logger
LOGGER = get_logger("Structure")

def get_feature(data):
    """return features : numpy.ndarray
    """
    for fk in 'x feat'.split():
        if fk in data.nodes.data:
            features=data.nodes.data[fk].numpy()
    return features

def get_edges(data):
    return data.edges.connections

def set_edges(data,adj):
    data.data["edge_index"]=adj

def to_adjacency_matrix(adj):
    """
    adj : torch.Tensor [2,E]
    return Tensor [N,N]
    """
    num_nodes=adj.max().item()+1
    mat = torch.zeros((num_nodes,num_nodes), dtype=bool)
    mat[tuple(adj)]=1
    return mat

def to_adjacency_list(adj):
    """
    adj : Tensor [N,N]
    return Tensor [2,E]
    """
    adj = torch.stack(adj.nonzero(as_tuple=True)).long() # edge list 
    return adj






import logging
import numpy as np
import torch
from tqdm import tqdm
import autogl
from ._basic import BaseFeatureGenerator
from ..._data_preprocessor_registry import DataPreprocessorUniversalRegistry

_LOGGER = logging.getLogger("FE")


class _Graphlet:
    def __init__(self, data, sample_error=0.1, sample_confidence=0.1):
        self._data = data
        self._init()

        self._sample_error = sample_error
        self._sample_confidence = sample_confidence
        self._dw = int(
            np.ceil(
                0.5 * (self._sample_error ** -2) * np.log(2 / self._sample_confidence)
            )
        )
        _LOGGER.info(
            "sample error {} , confidence {},num {}".format(
                self._sample_error, self._sample_confidence, self._dw
            )
        )

    def _init(self):
        self._edges = list(self._data.edge_index)
        self._edges = [self._edges[0], self._edges[1]]
        self._num_nodes = self._data.x.shape[0]
        self._num_edges = len(self._edges[0])
        self._neighbours = [[] for _ in range(self._num_nodes)]
        for i in range(len(self._edges[0])):
            u, v = self._edges[0][i], self._edges[1][i]
            self._neighbours[u].append(v)

        _LOGGER.info("nodes {} , edges {}".format(self._num_nodes, self._num_edges))

        # sorting
        self._node_degrees = np.array([len(x) for x in self._neighbours])
        self._nodes = np.argsort(self._node_degrees)
        for i in self._nodes:
            self._neighbours[i] = [
                x
                for _, x in sorted(
                    zip(self._node_degrees[self._neighbours[i]], self._neighbours[i]),
                    reverse=True,
                )
            ]
        self._neighbours = [np.array(x) for x in self._neighbours]

    def _get_gdv(self, v, u):
        if self._node_degrees[v] >= self._node_degrees[u]:
            pass
        else:
            u, v = v, u
        Sv, Su, Te = set(), set(), set()
        sigma1, sigma2 = 0, 0
        nb = self._neighbours
        N = self._num_nodes
        M = self._num_edges
        phi = np.zeros(self._num_nodes, dtype=int)
        c1, c2, c3, c4 = 1, 2, 3, 4
        x = np.zeros(16, dtype=int)
        # p1
        for w in nb[v]:
            if w != u:
                Sv.add(w)
                phi[w] = c1
        # p2
        for w in nb[u]:
            if w != v:
                if phi[w] == c1:
                    Te.add(w)
                    phi[w] = c3
                    Sv.remove(w)
                else:
                    Su.add(w)
                    phi[w] = c2
        # p3
        for w in Te:
            for r in nb[w]:
                if phi[r] == c3:
                    x[5] += 1
            phi[w] = c4
            sigma2 = sigma2 + len(nb[w]) - 2
        # p4
        for w in Su:
            for r in nb[w]:
                if phi[r] == c1:
                    x[8] += 1
                if phi[r] == c2:
                    x[7] += 1
                if phi[r] == c4:
                    sigma1 += 1
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1
        # p5
        for w in Sv:
            for r in nb[w]:
                if phi[r] == c1:
                    x[7] += 1
                if phi[r] == c4:
                    sigma1 += 1
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1

        lsv, lsu, lte, du, dv = len(Sv), len(Su), len(Te), len(nb[u]), len(nb[v])
        # 3-graphlet
        x[1] = lte
        x[2] = du + dv - 2 - 2 * x[1]
        x[3] = N - x[2] - x[1] - 2
        x[4] = N * (N - 1) * (N - 2) / 6 - (x[1] + x[2] + x[3])
        # 4 connected graphlets
        x[6] = x[1] * (x[1] - 1) / 2 - x[5]
        x[10] = lsv * lsu - x[8]
        x[9] = lsv * (lsv - 1) / 2 + lsu * (lsu - 1) / 2 - x[7]
        # 4 disconnected graphlets
        t1 = N - (lte + lsu + lsv + 2)
        x[11] = x[1] * t1
        x[12] = M - (du + dv - 1) - (sigma2 - sigma1 - x[5] - x[8] - x[7])
        x[13] = (lsu + lsv) * t1
        x[14] = t1 * (t1 - 1) / 2 - x[12]
        x[15] = N * (N - 1) * (N - 2) * (N - 3) / 24 - np.sum(x[5:15])

        return x

    def _get_gdv_sample(self, v, u):
        if self._node_degrees[v] >= self._node_degrees[u]:
            pass
        else:
            u, v = v, u
        Sv = set()
        sigma1, sigma2 = 0, 0
        nb = self._neighbours
        N = self._num_nodes
        M = self._num_edges
        phi = np.zeros(self._num_nodes, dtype=int)
        c1, c2, c3, c4 = 1, 2, 3, 4
        x = np.zeros(16)
        dw = self._dw

        # p1
        Sv = set(nb[v][nb[v] != u])
        phi[list(Sv)] = c1
        # p2
        p2w = nb[u][nb[u] != c1]
        p2w1 = p2w[phi[p2w] == c1]
        p2w2 = p2w[phi[p2w] != c1]
        Te = p2w1
        phi[p2w1] = c3
        Sv -= set(list(p2w1))
        Su = p2w2
        phi[p2w2] = c2
        # p3
        for w in Te:
            if dw >= len(nb[w]):
                region = nb[w]
                inc = 1
            else:
                region = np.random.choice(nb[w], dw, replace=False)
                inc = self._node_degrees[w] / dw
            phir = phi[region]
            x[5] += inc * np.sum(phir == c3)
            phi[w] = c4
            sigma2 = sigma2 + len(nb[w]) - 2
        # p4
        for w in Su:
            if dw >= len(nb[w]):
                region = nb[w]
                inc = 1
            else:
                region = np.random.choice(nb[w], dw, replace=False)
                inc = self._node_degrees[w] / dw
            phir = phi[region]
            x[8] += inc * np.sum(phir == c1)
            x[7] += inc * np.sum(phir == c2)
            sigma1 += inc * np.sum(phir == c4)
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1
        # p5
        for w in Sv:
            if dw >= len(nb[w]):
                region = nb[w]
                inc = 1
            else:
                region = np.random.choice(nb[w], dw, replace=False)
                inc = self._node_degrees[w] / dw
            phir = phi[region]
            x[7] += inc * np.sum(phir == c1)
            sigma1 += inc * np.sum(phir == c4)
            phi[w] = 0
            sigma2 = sigma2 + len(nb[w]) - 1

        lsv, lsu, lte, du, dv = len(Sv), len(Su), len(Te), len(nb[u]), len(nb[v])
        # 3-graphlet
        x[1] = lte
        x[2] = du + dv - 2 - 2 * x[1]
        x[3] = N - x[2] - x[1] - 2
        x[4] = N * (N - 1) * (N - 2) / 6 - (x[1] + x[2] + x[3])
        # 4 connected graphlets
        x[6] = x[1] * (x[1] - 1) / 2 - x[5]
        x[10] = lsv * lsu - x[8]
        x[9] = lsv * (lsv - 1) / 2 + lsu * (lsu - 1) / 2 - x[7]
        # 4 disconnected graphlets
        t1 = N - (lte + lsu + lsv + 2)
        x[11] = x[1] * t1
        x[12] = M - (du + dv - 1) - (sigma2 - sigma1 - x[5] - x[8] - x[7])
        x[13] = (lsu + lsv) * t1
        x[14] = t1 * (t1 - 1) / 2 - x[12]
        x[15] = N * (N - 1) * (N - 2) * (N - 3) / 24 - np.sum(x[5:15])

        return x

    def get_gdvs(self, sample=True):
        res = np.zeros((self._num_nodes, 15))
        for u in tqdm(range(self._num_nodes)):
            vs = self._neighbours[u]
            if len(vs) != 0:
                gdvs = []
                for v in tqdm(vs, disable=len(vs) < 100):
                    if sample:
                        gdvs.append(self._get_gdv_sample(u, v))
                    else:
                        gdvs.append(self._get_gdv(u, v))
                res[u, :] = np.mean(gdvs, axis=0)[1:]
        return res


@DataPreprocessorUniversalRegistry.register_data_preprocessor("graph" + "let")
class GraphletGenerator(BaseFeatureGenerator):
    r"""generate local graphlet numbers as features. The implementation refers to [#]_ .

    References
    ----------
    .. [#] Ahmed, N. K., Willke, T. L., & Rossi, R. A. (2016).
        Estimation of local subgraph counts. Proceedings - 2016 IEEE International Conference on Big Data, Big Data 2016, 586–595.
        https://doi.org/10.1109/BigData.2016.7840651

    """

    def _extract_nodes_feature(self, data: autogl.data.Data) -> torch.Tensor:
        result: np.ndarray = _Graphlet(data).get_gdvs()
        return torch.from_numpy(result)



import dgl
import torch.nn.functional
import typing as _typing
from dgl.nn.pytorch.conv import GraphConv
from .. import base_encoder, encoder_registry


class ApplyNodeFunc(torch.nn.Module):
    """Update the node feature hv with MLP, BN and ReLU."""
    def __init__(self, mlp):
        super(ApplyNodeFunc, self).__init__()
        self.mlp = mlp
        self.bn = torch.nn.BatchNorm1d(self.mlp.output_dim)

    def forward(self, h):
        h = self.mlp(h)
        h = self.bn(h)
        h = torch.nn.functional.relu(h)
        return h


class MLP(torch.nn.Module):
    """ MLP with linear output """
    def __init__(self, num_layers, input_dim, hidden_dim, output_dim):
        """MLP layers construction

        Paramters
        ---------
        num_layers: int
            The number of linear layers
        input_dim: int
            The dimensionality of input features
        hidden_dim: int
            The dimensionality of hidden units at ALL layers
        output_dim: int
            The number of classes for prediction

        """
        super(MLP, self).__init__()
        self.linear_or_not = True  # default is linear model
        self.num_layers = num_layers
        self.output_dim = output_dim

        if num_layers < 1:
            raise ValueError("number of layers should be positive!")
        elif num_layers == 1:
            # Linear model
            self.linear = torch.nn.Linear(input_dim, output_dim)
        else:
            # Multi-layer model
            self.linear_or_not = False
            self.linears = torch.nn.ModuleList()
            self.batch_norms = torch.nn.ModuleList()

            self.linears.append(torch.nn.Linear(input_dim, hidden_dim))
            for layer in range(num_layers - 2):
                self.linears.append(torch.nn.Linear(hidden_dim, hidden_dim))
            self.linears.append(torch.nn.Linear(hidden_dim, output_dim))

            for layer in range(num_layers - 1):
                self.batch_norms.append(torch.nn.BatchNorm1d(hidden_dim))

    def forward(self, x):
        if self.linear_or_not:
            # If linear model
            return self.linear(x)
        else:
            # If MLP
            h = x
            for i in range(self.num_layers - 1):
                h = torch.nn.functional.relu(self.batch_norms[i](self.linears[i](h)))
            return self.linears[-1](h)


class _TopK(torch.nn.Module):
    def __init__(self, input_dimension: int, dimensions: _typing.Sequence[int]):
        super(_TopK, self).__init__()
        self.__gcn_layers: torch.nn.ModuleList = torch.nn.ModuleList()
        self.__batch_normalizations: torch.nn.ModuleList = torch.nn.ModuleList()
        self.__num_layers = len(dimensions)
        for layer in range(self.__num_layers):
            self.__gcn_layers.append(
                GraphConv(
                    input_dimension if layer == 0 else dimensions[layer - 1],
                    dimensions[layer]
                )
            )
            self.__batch_normalizations.append(
                torch.nn.BatchNorm1d(dimensions[layer])
            )

    def forward(self, graph: dgl.DGLGraph, *__args, **__kwargs) -> _typing.Sequence[torch.Tensor]:
        graph = dgl.remove_self_loop(graph)
        graph = dgl.add_self_loop(graph)
        h = graph.ndata['feat']
        hidden_rep = [h]
        for i in range(self.__num_layers):
            h = self.__gcn_layers[i](graph, h)
            h = self.__batch_normalizations[i](h)
            h = torch.nn.functional.relu(h)
            hidden_rep.append(h)
        return hidden_rep


@encoder_registry.EncoderUniversalRegistry.register_encoder('TopK'.lower())
@encoder_registry.EncoderUniversalRegistry.register_encoder('TopK_encoder'.lower())
class AutoTopKMaintainer(base_encoder.AutoHomogeneousEncoderMaintainer):
    def __init__(
            self,
            input_dimension: _typing.Optional[int] = ...,
            final_dimension: _typing.Optional[int] = ...,
            device: _typing.Union[torch.device, str, int, None] = ...,
            *args, **kwargs
    ):
        super(AutoTopKMaintainer, self).__init__(
            input_dimension, final_dimension,
            device, *args, **kwargs
        )
        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "4,5,6",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 5,
                "minValue": [8, 8, 8, 8, 8],
                "maxValue": [64, 64, 64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            }
        ]
        self.hyper_parameters = {
            "num_layers": 5,
            "hidden": [64, 64, 64, 64]
        }

    def _initialize(self) -> _typing.Optional[bool]:
        dimensions = list(self.hyper_parameters["hidden"])
        if (
                self.final_dimension not in (Ellipsis, None) and
                isinstance(self.final_dimension, int) and
                self.final_dimension > 0
        ):
            dimensions.append(self.final_dimension)
        self._encoder = _TopK(
            self.input_dimension, dimensions
        ).to(self.device)
        return True



import numpy as np
import torch


def data_is_tensor(data):
    return isinstance(data.x, torch.Tensor)


def data_is_numpy(data):
    return isinstance(data.x, np.ndarray)


def data_tensor2np(data):
    if data_is_tensor(data):
        data.x = data.x.numpy()
        data.y = data.y.numpy()
        data.edge_index = data.edge_index.numpy()
    return data


def data_np2tensor(data):
    if not data_is_tensor(data):
        if data_is_numpy(data):
            data.x = torch.FloatTensor(data.x)
        data.y = torch.tensor(data.y)
        data.edge_index = torch.tensor(data.edge_index, dtype=torch.long)
    return data


# from .base import BaseFeatureAtom
# class DataTensor2Np(BaseFeatureAtom):
#     def __call__(self,data):
#         return data_tensor2np(data)
# class DataNp2Tensor(BaseFeatureAtom):
#     def __call__(self,data):
#         return data_np2tensor(data)



"""
HPO Module for tuning hyper parameters
"""

from ...utils import get_logger
import hyperopt
import time
import math
import dill
import multiprocessing as mp
import copy
import random

mp.set_start_method("spawn", True)


LOGGER = get_logger("HPO")


class BaseHPOptimizer:
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.logger = LOGGER

    def _decompose_dld(self, config):
        self._dld = {}
        if isinstance(config, list):
            return config
        # config is dict
        list_config = []
        for key in config:
            self._dld[key] = []
            for para in config[key]:
                self._dld[key].append(para["parameterName"])
                newpara = para.copy()
                newpara["parameterName"] = key + ":" + para["parameterName"]
                if "cutPara" in para.keys():
                    if isinstance(newpara["cutPara"], str):
                        newpara["cutPara"] = key + ":" + para["cutPara"]
                    else:
                        newpara["cutPara"] = [key + ":" + cutname for cutname in para["cutPara"]]
                list_config.append(newpara)
        return list_config

    def _compose_dld(self, para):
        if not self._dld:
            return para
        fin = {}
        for key in self._dld:
            fin[key] = {}
            for pname in self._dld[key]:
                fin[key][pname] = para[key + ":" + pname]
        return fin

    def _decompose_depend_list_para(self, config):
        self._depend_map = {}

        def get_depended_para(name):
            for p in config:
                if p["parameterName"] == name:
                    return
            raise WrongDependedParameterError("The depended parameter does not exist.")

        for para in config:
            if para["type"] in ("NUMERICAL_LIST", "CATEGORICAL_LIST") and para.get(
                "cutPara", None
            ):
                self._depend_map[para["parameterName"]] = para
                if type(para["cutPara"]) == str:
                    get_depended_para(para["cutPara"])
                else:
                    for dpara in para["cutPara"]:
                        get_depended_para(dpara)

        return config

    def _compose_depend_list_para(self, config):
        for para in self._depend_map:
            cutparas = self._depend_map[para]["cutPara"]
            if type(cutparas) == str:
                dparas = [config[cutparas]]
                # dparas = config[cutparas]
            else:
                dparas = []
                for dpara in cutparas:
                    dparas.append(config[dpara])
            paralen = self._depend_map[para]["cutFunc"](dparas)
            config[para] = config[para][:paralen]
        return config

    def _decompose_list_fixed_para(self, config):
        fin = []
        self._list_map = {}
        self._fix_map = {}
        for para in config:
            if para["type"] == "NUMERICAL_LIST":
                self._list_map[para["parameterName"]] = para["length"]
                if type(para["minValue"]) != list:
                    para["minValue"] = [para["minValue"] for i in range(para["length"])]
                if type(para["maxValue"]) != list:
                    para["maxValue"] = [para["maxValue"] for i in range(para["length"])]
                for i, x, y in zip(
                    range(para["length"]), para["minValue"], para["maxValue"]
                ):
                    new_para = {}
                    new_para["parameterName"] = para["parameterName"] + "_" + str(i)
                    new_para["type"] = para["numericalType"]
                    new_para["minValue"] = x
                    new_para["maxValue"] = y
                    new_para["scalingType"] = para["scalingType"]
                    fin.append(new_para)
            elif para["type"] == "CATEGORICAL_LIST":
                self._list_map[para["parameterName"]] = para["length"]
                category = para["feasiblePoints"]
                self._category_map[para["parameterName"]] = category

                cur_points = ",".join(map(lambda _x: str(_x), range(len(category))))
                for i in range(para["length"]):
                    new_para = dict()
                    new_para["parameterName"] = para["parameterName"] + "_" + str(i)
                    new_para["type"] = "DISCRETE"
                    new_para["feasiblePoints"] = cur_points
                    fin.append(new_para)
            elif para["type"] == "FIXED":
                self._fix_map[para["parameterName"]] = para["value"]
            else:
                new_para = para.copy()
                new_para["parameterName"] = para["parameterName"] + "_"
                fin.append(new_para)
        return fin

    def _compose_list_fixed_para(self, config):
        fin = {}
        # compose list
        for pname in self._list_map:
            val = []
            for i in range(self._list_map[pname]):
                val.append(config[pname + "_" + str(i)])
                del config[pname + "_" + str(i)]
            if pname in self._category_map:
                val = [self._category_map[pname][i] for i in val]
            fin[pname] = val
        # deal other para
        for pname in config:
            fin[pname[:-1]] = config[pname]
        for pname in self._fix_map:
            fin[pname] = self._fix_map[pname]
        return fin

    def _encode_para_convert(self, config):
        """
        Convert all types of para space to DOUBLE(linear), DISCRETE
        config: [{
            "parameterName": "num_layers",
            "type": "DISCRETE",
            "feasiblePoints": "1,2,3,4",
        },{
            "parameterName": "hidden",
            "type": "NUMERICAL_LIST",
            "numericalType": "INTEGER",
            "length": 4,
            "minValue": [4, 4, 4, 4],
            "maxValue": [32, 32, 32, 32],
            "scalingType": "LOG"
        },{
            "parameterName": "dropout",
            "type": "DOUBLE",
            "minValue": 0.1,
            "maxValue": 0.9,
            "scalingType": "LINEAR"
        }]"""
        self._category_map = {}
        self._discrete_map = {}
        self._numerical_map = {}

        current_config = []
        for para in config:
            if para["type"] == "DOUBLE" or para["type"] == "INTEGER":
                cur_para = para.copy()
                cur_para["type"] = "DOUBLE"
                if para["scalingType"] == "LOG":
                    cur_para["minValue"] = math.log(para["minValue"])
                    cur_para["maxValue"] = math.log(para["maxValue"])
                current_config.append(cur_para)
                self._numerical_map[para["parameterName"]] = para
            elif para["type"] == "CATEGORICAL" or para["type"] == "DISCRETE":
                if para["type"] == "DISCRETE":
                    cate_list = para["feasiblePoints"].split(",")
                    cate_list = list(map(lambda x: x.strip(), cate_list))
                else:
                    cate_list = para["feasiblePoints"]
                cur_points = ",".join(map(lambda x: str(x), range(len(cate_list))))
                cur_para = para.copy()
                cur_para["feasiblePoints"] = cur_points
                cur_para["type"] = "DISCRETE"
                current_config.append(cur_para)
                if para["type"] == "CATEGORICAL":
                    self._category_map[para["parameterName"]] = cate_list
                else:
                    self._discrete_map[para["parameterName"]] = cate_list
            else:
                current_config.append(para)
        return current_config

    def _decode_para_convert(self, para):
        """
        decode HPO given para to user(externel) para and trial para
        """
        externel_para = para.copy()
        trial_para = para.copy()
        for name in para:
            if name in self._numerical_map:
                old_para = self._numerical_map[name]
                val = para[name]
                if old_para["scalingType"] == "LOG":
                    val = math.exp(val)
                    if val < old_para["minValue"]:
                        val = old_para["minValue"]
                    elif val > old_para["maxValue"]:
                        val = old_para["maxValue"]
                if old_para["type"] == "INTEGER":
                    val = int(round(val))
                externel_para[name] = val
                trial_para[name] = (
                    val if old_para["scalingType"] != "LOG" else math.log(val)
                )
            elif name in self._category_map:
                externel_para[name] = self._category_map[name][int(para[name])]
                trial_para[name] = para[name]
            elif name in self._discrete_map:
                externel_para[name] = eval(self._discrete_map[name][int(para[name])])
                trial_para[name] = para[name]
        return externel_para, trial_para

    def _encode_para(self, config):
        # dict(list(dict)) -> list(dict)
        config = self._decompose_dld(config)
        # (dependent list) -> list
        config = self._decompose_depend_list_para(config)
        # list -> double/discrete; fixed -> removed
        config = self._decompose_list_fixed_para(config)
        # discrete -> categorical; log -> linear
        config = self._encode_para_convert(config)
        return config

    def _decode_para(self, para):
        para, trial_para = self._decode_para_convert(para)
        para = self._compose_list_fixed_para(para)
        para = self._compose_depend_list_para(para)
        para = self._compose_dld(para)
        return para, trial_para

    def _print_info(self, para, perf):
        if self.is_higher_better:
            # print("Parameter: {} {}: {} {}".format(para, trainer.get_feval(return_major=True).get_eval_name(), -perf, "higher_better"))
            LOGGER.info(
                "Parameter: {} {}: {} {}".format(
                    para, self.feval_name, -perf, "higher_better"
                )
            )
        else:
            # print("Parameter: {} {}: {} {}".format(para, trainer.get_feval(return_major=True).get_eval_name(), perf, "lower_better"))
            LOGGER.info(
                "Parameter: {} {}: {} {}".format(
                    para, self.feval_name, perf, "lower_better"
                )
            )

    def slave(self, pipe, trainer, dataset):
        while 1:
            x = pipe.recv()
            current_trainer = trainer.duplicate_from_hyper_parameter(x)
            current_trainer.train(dataset)
            loss, is_higher_better = current_trainer.get_valid_score(dataset)
            trainer_pic = dill.dumps(current_trainer)
            pipe.send(trainer_pic)
            pipe.send(loss)

    def _set_up(self, trainer, dataset, time_limit, memory_limit):
        """
        Initialize something used in "optimize"

        Parameters
        ----------
        trainer : ..train.BaseTrainer
            Including model, giving HP space and using for training
        dataset : ...datasets
            Dataset to train and evaluate.
        time_limit : int
            Max time to run HPO
        memory_limit : None
            No implementation yet
        """
        pass

    def _update_trials(self, pid, cur_trainer, perf, is_higher_better):
        """
        After the evaluation phase of each turn, update history trials according to the performance

        Parameters
        ----------
        pid : int
            The process id which runs the evaluation
        cur_trainer : ..train.BaseTrainer
            The trainer used in evaluation, containing the HPs
        perf : float
            The performance of the HP
        is_higher_better : bool
            True if higher perf is better
        """
        pass

    def _get_suggestion(self, pid):
        """
        Give the next HP suggestion

        Parameters
        ----------
        pid : int
            The process id which will run the evaluation

        Returns
        -------
        para_json: dict
            The suggested HP
        """
        pass

    def _best_hp(self):
        """
        Give the best HP and the best trainer as the returns of "optimize"

        Returns
        -------
        trainer: ..train.BaseTrainer
            The trainer including the best trained model
        para_json: dict
            The best HP
        """
        pass

    def optimize(self, trainer, dataset, time_limit=None, memory_limit=None):
        """
        The main process of optimizing the HP by the method within give model and HP space.
        This process provide an HPO scheme, where giving suggestion and evaluation are taken turns, and can be run on multi-GPU.
        If you want to implement your HPO algorithm, you can either follow this scheme, you should only change "_set_up", "_update_trials", "_get_suggestion", and "_best_hp".
        Or you can rewrite the whole "optimize" function.
        Parameters
        ----------
        trainer : ..train.BaseTrainer
            Including model, giving HP space and using for training
        dataset : ...datasets
            Dataset to train and evaluate.
        time_limit : int
            Max time to run HPO
        memory_limit : None
            No implementation yet

        Returns
        -------
        trainer: ..train.BaseTrainer
            The trainer including the best trained model
        para_json: dict
            The best HP
        """
        start_time = time.time()
        cur_evals = 0

        slaves = 2
        pipes = []
        procs = []
        statuses = []
        for i in range(slaves):
            ps, pr = mp.Pipe()
            pipes.append(ps)
            proc = mp.Process(
                target=self.slave,
                args=(
                    pr,
                    trainer,
                    dataset,
                ),
            )
            proc.start()
            procs.append(proc)
            statuses.append("ready")

        # do something of the certain HPO algo
        self._set_up(trainer, dataset, slaves, time_limit, memory_limit)
        self.feval_name = trainer.get_feval(return_major=True).get_eval_name()
        self.is_higher_better = trainer.get_feval(return_major=True).is_higher_better()

        while cur_evals < self.max_evals:
            # timeout
            if time.time() - start_time > time_limit:
                self.logger.info("Time out of limit, Epoch: {}".format(str(cur_evals)))
                break
            for i in range(slaves):
                sent = pipes[i].poll()
                if sent or statuses[i] == "ready":
                    if sent:
                        cur_trainer = pipes[i].recv()
                        cur_trainer = dill.loads(cur_trainer)
                        perf = pipes[i].recv()
                        self._update_trials(i, cur_trainer, perf, self.is_higher_better)
                        cur_evals += 1
                        if cur_evals >= self.max_evals:
                            break
                    else:
                        statuses[i] = "running"
                    x = self._get_suggestion(i)
                    pipes[i].send(x)

        for i in range(slaves):
            procs[i].terminate()

        best_trainer, best_x = self._best_hp()
        if best_x == None:
            raise TimeTooLimitedError(
                "Given time is too limited to finish one round in HPO."
            )
        return best_trainer, best_x

    @classmethod
    def build_hpo_from_args(cls, args):
        """Build a new hpo instance."""
        raise NotImplementedError(
            "HP Optimizer must implement the build_hpo_from_args method"
        )


class TimeTooLimitedError(Exception):
    pass


class WrongDependedParameterError(Exception):
    pass

def test():
    config = {"encoder": [{
        "parameterName": "num_layers",
        "type": "DISCRETE",
        "feasiblePoints": "1,2,3,4",
    },{
        "parameterName": "hidden",
        "type": "NUMERICAL_LIST",
        "numericalType": "INTEGER",
        "length": 4,
        "cutPara": "num_layers",
        "cutFunc": lambda x: x[0],
        "minValue": [4, 4, 4, 4],
        "maxValue": [32, 32, 32, 32],
        "scalingType": "LOG"
    },{
        "parameterName": "dropout",
        "type": "DOUBLE",
        "minValue": 0.1,
        "maxValue": 0.9,
        "scalingType": "LINEAR"
    }],
    "decoder":[{
        "parameterName": "dropout",
        "type": "DOUBLE",
        "minValue": 0.1,
        "maxValue": 0.9,
        "scalingType": "LINEAR"
    }]}
    bhpo = BaseHPOptimizer()
    newconf = bhpo._encode_para(config)
    print(newconf)

    para = {}
    for i in newconf:
        if i['type'] == "DISCRETE":
            para[i['parameterName']] = int(random.choice(i["feasiblePoints"].split(',')))
        else:
            para[i['parameterName']] = i['minValue'] + random.random() * (i['maxValue'] - i["minValue"])
    newpara = bhpo._decode_para(para)[0]
    print(newpara)

if __name__ == "__main__":
    test()
     



from scipy.sparse import csr_matrix
import scipy.sparse as ssp
import networkx as nx
import numpy as np

# @register_feature('pagerank')
from .base import BaseGenerator
from .. import register_feature


@register_feature("pagerank")
class GePageRank(BaseGenerator):
    r"""concat pagerank features"""

    def _transform(self, data):
        graph = nx.DiGraph()
        w = data.edge_weight
        eg = [(u, v, w[i]) for i, (u, v) in enumerate(data.edge_index.T)]
        graph.add_weighted_edges_from(eg)
        pagerank = nx.pagerank(graph)
        pr = np.zeros((data.x.shape[0], 1))
        for i, v in pagerank.items():
            pr[i] = v
        data.x = np.concatenate([data.x, pr], axis=1)
        return data

    def _preprocess(self, data):
        if not hasattr(data, "edge_weight"):
            data.edge_weight = np.ones(data.edge_index.shape[1])

    def _postprocess(self, data):
        del data.edge_weight



"""
NAS algorithms
"""

import importlib
import os
from .base import BaseNAS

NAS_ALGO_DICT = {}


def register_nas_algo(name):
    def register_nas_algo_cls(cls):
        if name in NAS_ALGO_DICT:
            raise ValueError(
                "Cannot register duplicate NAS algorithm ({})".format(name)
            )
        if not issubclass(cls, BaseNAS):
            raise ValueError(
                "Model ({}: {}) must extend NAS algorithm".format(name, cls.__name__)
            )
        NAS_ALGO_DICT[name] = cls
        return cls

    return register_nas_algo_cls


from .darts import Darts
from .enas import Enas
from .random_search import RandomSearch
from .rl import RL, GraphNasRL
from ..backend import *
if not is_dgl():
    from .gasso import Gasso
    from .grna import GRNA
from .spos import Spos

def build_nas_algo_from_name(name: str) -> BaseNAS:
    """
    Parameters
    ----------
    name: ``str``
        the name of nas algorithm.

    Returns
    -------
    BaseNAS:
        the NAS algorithm built using default parameters

    Raises
    ------
    AssertionError
        If an invalid name is passed in
    """
    assert name in NAS_ALGO_DICT, "HPO module do not have name " + name
    return NAS_ALGO_DICT[name]()


__all__ = ["BaseNAS", "Darts", "Enas", "RandomSearch", "RL", "GraphNasRL","Spos"]
if not is_dgl():
    __all__.append("Gasso")
    __all__.append("GRNA")



import typing as _typing
import torch
import torch_geometric
from autogl.data import Dataset, InMemoryDataset
from autogl.data.graph import GeneralStaticGraph
from autogl.data.graph.utils import conversion


def to_pyg_dataset(
        dataset: _typing.Union[Dataset, _typing.Iterable[GeneralStaticGraph]]
) -> Dataset[torch_geometric.data.Data]:
    transformed_datalist: _typing.MutableSequence[torch_geometric.data.Data] = []
    for item in dataset:
        if isinstance(item, torch_geometric.data.Data):
            transformed_datalist.append(item)
        elif isinstance(item, GeneralStaticGraph):
            transformed_datalist.append(conversion.static_graph_to_pyg_data(item))
        elif (
                isinstance(item, _typing.Mapping) and
                all([
                    (isinstance(k, str) and isinstance(v, torch.Tensor))
                    for k, v in item.items()
                ])
        ):
            transformed_datalist.append(torch_geometric.data.Data(**item))
        else:
            raise NotImplementedError(
                f"Unsupported data item {type(item)}<{item}> to convert as "
                f"{torch_geometric.data.Data}"
            )
    return (
        InMemoryDataset(transformed_datalist, dataset.train_index, dataset.val_index, dataset.test_index, dataset.schema)
        if isinstance(dataset, InMemoryDataset)
        else InMemoryDataset(transformed_datalist)
    )



import torch
import typing as _typing

from torch_geometric.nn.conv import SAGEConv
import torch.nn.functional
import autogl.data
from . import register_model
from .base import BaseAutoModel, activate_func, ClassificationSupportedSequentialModel
from ....utils import get_logger

LOGGER = get_logger("SAGEModel")


class GraphSAGE(ClassificationSupportedSequentialModel):
    class _SAGELayer(torch.nn.Module):
        def __init__(
            self,
            input_channels: int,
            output_channels: int,
            aggr: str,
            activation_name: _typing.Optional[str] = ...,
            dropout_probability: _typing.Optional[float] = ...,
        ):
            super().__init__()
            self._convolution: SAGEConv = SAGEConv(
                input_channels, output_channels, aggr=aggr
            )
            if (
                activation_name is not Ellipsis
                and activation_name is not None
                and type(activation_name) == str
            ):
                self._activation_name: _typing.Optional[str] = activation_name
            else:
                self._activation_name: _typing.Optional[str] = None
            if (
                dropout_probability is not Ellipsis
                and dropout_probability is not None
                and type(dropout_probability) == float
            ):
                if dropout_probability < 0:
                    dropout_probability = 0
                if dropout_probability > 1:
                    dropout_probability = 1
                self._dropout: _typing.Optional[torch.nn.Dropout] = torch.nn.Dropout(
                    dropout_probability
                )
            else:
                self._dropout: _typing.Optional[torch.nn.Dropout] = None

        def forward(self, data, enable_activation: bool = True) -> torch.Tensor:
            x: torch.Tensor = getattr(data, "x")
            edge_index: torch.Tensor = getattr(data, "edge_index")
            if type(x) != torch.Tensor or type(edge_index) != torch.Tensor:
                raise TypeError

            x: torch.Tensor = self._convolution.forward(x, edge_index)
            if self._activation_name is not None and enable_activation:
                x: torch.Tensor = activate_func(x, self._activation_name)
            if self._dropout is not None:
                x: torch.Tensor = self._dropout.forward(x)
            return x

    def __init__(
        self,
        num_features: int,
        num_classes: int,
        hidden_features: _typing.Sequence[int],
        activation_name: str,
        layers_dropout: _typing.Union[
            _typing.Optional[float], _typing.Sequence[_typing.Optional[float]]
        ] = None,
        aggr: str = "mean",
    ):
        super().__init__()
        if not type(num_features) == type(num_classes) == int:
            raise TypeError
        if not isinstance(hidden_features, _typing.Sequence):
            raise TypeError
        for hidden_feature in hidden_features:
            if type(hidden_feature) != int:
                raise TypeError
            elif hidden_feature <= 0:
                raise ValueError
        if isinstance(layers_dropout, _typing.Sequence):
            if len(layers_dropout) != (len(hidden_features) + 1):
                raise TypeError
            for d in layers_dropout:
                if d is not None and type(d) != float:
                    raise TypeError
            _layers_dropout: _typing.Sequence[_typing.Optional[float]] = layers_dropout
        elif layers_dropout is None or type(layers_dropout) == float:
            _layers_dropout: _typing.Sequence[_typing.Optional[float]] = [
                layers_dropout for _ in range(len(hidden_features))
            ] + [None]
        else:
            raise TypeError
        if not type(activation_name) == type(aggr) == str:
            raise TypeError
        if aggr not in ("add", "max", "mean"):
            aggr = "mean"

        if len(hidden_features) == 0:
            self.__sequential_encoding_layers: torch.nn.ModuleList = (
                torch.nn.ModuleList(
                    [
                        self._SAGELayer(
                            num_features,
                            num_classes,
                            aggr,
                            activation_name,
                            _layers_dropout[0],
                        )
                    ]
                )
            )
        else:
            self.__sequential_encoding_layers: torch.nn.ModuleList = (
                torch.nn.ModuleList(
                    [
                        self._SAGELayer(
                            num_features,
                            hidden_features[0],
                            aggr,
                            activation_name,
                            _layers_dropout[0],
                        )
                    ]
                )
            )
            for i in range(len(hidden_features)):
                if i + 1 < len(hidden_features):
                    self.__sequential_encoding_layers.append(
                        self._SAGELayer(
                            hidden_features[i],
                            hidden_features[i + 1],
                            aggr,
                            activation_name,
                            _layers_dropout[i + 1],
                        )
                    )
                else:
                    self.__sequential_encoding_layers.append(
                        self._SAGELayer(
                            hidden_features[i],
                            num_classes,
                            aggr,
                            _layers_dropout[i + 1],
                        )
                    )

    @property
    def sequential_encoding_layers(self) -> torch.nn.ModuleList:
        return self.__sequential_encoding_layers

    def cls_encode(self, data) -> torch.Tensor:
        if (
            hasattr(data, "edge_indexes")
            and isinstance(getattr(data, "edge_indexes"), _typing.Sequence)
            and len(getattr(data, "edge_indexes"))
            == len(self.__sequential_encoding_layers)
        ):
            for __edge_index in getattr(data, "edge_indexes"):
                if type(__edge_index) != torch.Tensor:
                    raise TypeError
            """ Layer-wise encode """
            x: torch.Tensor = getattr(data, "x")
            for i, __edge_index in enumerate(getattr(data, "edge_indexes")):
                x: torch.Tensor = self.__sequential_encoding_layers[i](
                    autogl.data.Data(x=x, edge_index=__edge_index)
                )
            return x
        else:
            x: torch.Tensor = getattr(data, "x")
            for i in range(len(self.__sequential_encoding_layers)):
                x = self.__sequential_encoding_layers[i](
                    autogl.data.Data(x, getattr(data, "edge_index"))
                )
            return x

    def cls_decode(self, x: torch.Tensor) -> torch.Tensor:
        return torch.nn.functional.log_softmax(x, dim=1)

    def lp_encode(self, data):
        x: torch.Tensor = getattr(data, "x")
        for i in range(len(self.__sequential_encoding_layers) - 2):
            x = self.__sequential_encoding_layers[i](
                autogl.data.Data(x, getattr(data, "edge_index"))
            )
        x = self.__sequential_encoding_layers[-2](
            autogl.data.Data(x, getattr(data, "edge_index")), enable_activation=False
        )
        return x

    def lp_decode(self, z, pos_edge_index, neg_edge_index):
        edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
        logits = (z[edge_index[0]] * z[edge_index[1]]).sum(dim=-1)
        return logits

    def lp_decode_all(self, z):
        prob_adj = z @ z.t()
        return (prob_adj > 0).nonzero(as_tuple=False).t()


@register_model("sage-model")
class AutoSAGE(BaseAutoModel):
    r"""
    AutoSAGE. The model used in this automodel is GraphSAGE, i.e., the GraphSAGE from the `"Inductive Representation Learning on
    Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper. The layer is

    .. math::

        \mathbf{x}^{\prime}_i = \mathbf{W}_1 \mathbf{x}_i + \mathbf{W_2} \cdot
        \mathrm{mean}_{j \in \mathcal{N(i)}} \mathbf{x}_j

    Parameters
    ----------
    num_features: `int`.
        The dimension of features.

    num_classes: `int`.
        The number of classes.

    device: `torch.device` or `str`
        The device where model will be running on.

    init: `bool`.
        If True(False), the model will (not) be initialized.

    """

    def __init__(
        self, num_features=None, num_classes=None, device=None, **args
    ):

        super().__init__(num_features, num_classes, device, **args)

        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 3,
                "minValue": [8, 8, 8],
                "maxValue": [128, 128, 128],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.8,
                "minValue": 0.2,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "act",
                "type": "CATEGORICAL",
                "feasiblePoints": ["leaky_relu", "relu", "elu", "tanh"],
            },
            {
                "parameterName": "agg",
                "type": "CATEGORICAL",
                "feasiblePoints": ["mean", "add", "max"],
            },
        ]

        self.hyper_parameters = {
            "num_layers": 3,
            "hidden": [64, 32],
            "dropout": 0.5,
            "act": "relu",
            "agg": "mean",
        }


    def _initialize(self):
        self._model = GraphSAGE(
            self.input_dimension,
            self.output_dimension,
            self.hyper_parameters.get("hidden"),
            self.hyper_parameters.get("act", "relu"),
            self.hyper_parameters.get("dropout", None),
            self.hyper_parameters.get("agg", "mean"),
        ).to(self.device)



import torch
import torch.nn as nn
import torch.nn.functional as F
from dgl.nn.pytorch.conv import GraphConv
from dgl.nn.pytorch.glob import SortPooling
from . import register_model
from .base import BaseAutoModel
from ....utils import get_logger

LOGGER = get_logger("TopkModel")


def set_default(args, d):
    for k, v in d.items():
        if k not in args:
            args[k] = v
    return args


class ApplyNodeFunc(nn.Module):
    """Update the node feature hv with MLP, BN and ReLU."""
    def __init__(self, mlp):
        super(ApplyNodeFunc, self).__init__()
        self.mlp = mlp
        self.bn = nn.BatchNorm1d(self.mlp.output_dim)

    def forward(self, h):
        h = self.mlp(h)
        h = self.bn(h)
        h = F.relu(h)
        return h


class MLP(nn.Module):
    """MLP with linear output"""
    def __init__(self, num_layers, input_dim, hidden_dim, output_dim):
        """MLP layers construction

        Paramters
        ---------
        num_layers: int
            The number of linear layers
        input_dim: int
            The dimensionality of input features
        hidden_dim: int
            The dimensionality of hidden units at ALL layers
        output_dim: int
            The number of classes for prediction

        """
        super(MLP, self).__init__()
        self.linear_or_not = True  # default is linear model
        self.num_layers = num_layers
        self.output_dim = output_dim

        if num_layers < 1:
            raise ValueError("number of layers should be positive!")
        elif num_layers == 1:
            # Linear model
            self.linear = nn.Linear(input_dim, output_dim)
        else:
            # Multi-layer model
            self.linear_or_not = False
            self.linears = torch.nn.ModuleList()
            self.batch_norms = torch.nn.ModuleList()

            self.linears.append(nn.Linear(input_dim, hidden_dim))
            for layer in range(num_layers - 2):
                self.linears.append(nn.Linear(hidden_dim, hidden_dim))
            self.linears.append(nn.Linear(hidden_dim, output_dim))

            for layer in range(num_layers - 1):
                self.batch_norms.append(nn.BatchNorm1d((hidden_dim)))

    def forward(self, x):
        if self.linear_or_not:
            # If linear model
            return self.linear(x)
        else:
            # If MLP
            h = x
            for i in range(self.num_layers - 1):
                h = F.relu(self.batch_norms[i](self.linears[i](h)))
            return self.linears[-1](h)



class Topkpool(torch.nn.Module):
    """Topkpool model"""
    def __init__(self, args):
        """model parameters setting

        Paramters
        ---------
        num_layers: int
            The number of linear layers in the neural network
        num_mlp_layers: int
            The number of linear layers in mlps
        input_dim: int
            The dimensionality of input features
        hidden_dim: int
            The dimensionality of hidden units at ALL layers
        output_dim: int
            The number of classes for prediction
        final_dropout: float
            dropout ratio on the final linear layer

        """
        super(Topkpool, self).__init__()
        self.args = args

        missing_keys = list(
            set(
                [
                    "features_num",
                    "num_class",
                    "num_graph_features",
                    "num_layers",
                    "hidden",
                    "dropout",
                ]
            )
            - set(self.args.keys())
        )
        if len(missing_keys) > 0:
            raise Exception("Missing keys: %s." % ",".join(missing_keys))
        #if not self.num_layer == len(self.args["hidden"]) + 1:
        #    LOGGER.warn("Warning: layer size does not match the length of hidden units")


        self.num_graph_features = self.args["num_graph_features"]
        self.num_layers = self.args["num_layers"]
        assert self.num_layers > 2, "Number of layers in GIN should not less than 3"
        assert self.num_layers == len(self.args["hidden"]) + 1, "Warning: layer size does not match the length of hidden units"

        input_dim = self.args["features_num"]
        hidden = self.args["hidden"]
        final_dropout = self.args["dropout"]
        output_dim = self.args["num_class"]

        # List of MLPs
        self.gcnlayers = torch.nn.ModuleList()
        self.batch_norms = torch.nn.ModuleList()

        for layer in range(self.num_layers - 1):
            if layer == 0:
                self.gcnlayers.append(GraphConv(input_dim, hidden[layer]))
            else:
                self.gcnlayers.append(GraphConv(hidden[layer-1], hidden[layer]))

            #self.gcnlayers.append(GraphConv(input_dim, hidden_dim))
            self.batch_norms.append(nn.BatchNorm1d(hidden[layer]))

        # Linear function for graph poolings of output of each layer
        # which maps the output of different layers into a prediction score
        self.linears_prediction = torch.nn.ModuleList()

        #TopKPool
        k = 3
        self.pool = SortPooling(k)

        for layer in range(self.num_layers):
            if layer == 0:
                self.linears_prediction.append(
                    nn.Linear(input_dim * k, output_dim))
            else:
                self.linears_prediction.append(
                    nn.Linear(hidden[layer-1] * k, output_dim))

        self.drop = nn.Dropout(final_dropout)


    #def forward(self, g, h):
    def forward(self, data):
        h = data.ndata.pop('feat')
        # list of hidden representation at each layer (including input)
        hidden_rep = [h]

        for i in range(self.num_layers - 1):
            h = self.gcnlayers[i](data, h)
            h = self.batch_norms[i](h)
            h = F.relu(h)
            hidden_rep.append(h)

        score_over_layer = 0

        # perform pooling over all nodes in each graph in every layer
        for i, h in enumerate(hidden_rep):
            pooled_h = self.pool(data, h)
            #import pdb; pdb.set_trace()
            score_over_layer += self.drop(self.linears_prediction[i](pooled_h))

        return score_over_layer


@register_model("topkpool-model")
class AutoTopkpool(BaseAutoModel):
    r"""
    AutoTopkpool. The model used in this automodel is from https://arxiv.org/abs/1905.05178, https://arxiv.org/abs/1905.02850
    Parameters
    ----------
    num_features: `int`.
        The dimension of features.
    num_classes: `int`.
        The number of classes.
    device: `torch.device` or `str`
        The device where model will be running on.
    init: `bool`.
        If True(False), the model will (not) be initialized.
    """

    def __init__(
        self,
        num_features=None,
        num_classes=None,
        device=None,
        num_graph_features=0,
        **args
    ):
        super().__init__(num_features, num_classes, device, num_graph_features=num_graph_features)
        self.num_graph_features = num_graph_features

        self.hyper_parameter_space = [
            {
                "parameterName": "num_layers",
                "type": "DISCRETE",
                "feasiblePoints": "4,5,6",
            },
            {
                "parameterName": "hidden",
                "type": "NUMERICAL_LIST",
                "numericalType": "INTEGER",
                "length": 5,
                "minValue": [8, 8, 8, 8, 8],
                "maxValue": [64, 64, 64, 64, 64],
                "scalingType": "LOG",
                "cutPara": ("num_layers",),
                "cutFunc": lambda x: x[0] - 1,
            },
            {
                "parameterName": "dropout",
                "type": "DOUBLE",
                "maxValue": 0.9,
                "minValue": 0.1,
                "scalingType": "LINEAR",
            },
            {
                "parameterName": "num_layers",
                "type": "INTEGER",
                "minValue": 7,
                "maxValue": 2,
                "scalingType": "LINEAR"
            },
            {
                "parameterName": "mlp_layers",
                "type": "DISCRETE",
                "feasiblePoints": "2,3,4",
            },
        ]

        self.hyper_parameters = {
            "num_layers": 5,
            "hidden": [64,64,64,64],
            "dropout": 0.5,
            "act": "relu",
            "mlp_layers": 2
        }

    def from_hyper_parameter(self, hp, **kwargs):
        return super().from_hyper_parameter(hp, num_graph_features=self.num_graph_features, **kwargs)

    def _initialize(self):
        self._model = Topkpool({
            "features_num": self.input_dimension,
            "num_class": self.output_dimension,
            "num_graph_features": self.num_graph_features,
            **self.hyper_parameters
        }).to(self.device)




import typing as _typing
from .base import BaseAutoModel

MODEL_DICT: _typing.Dict[str, _typing.Type[BaseAutoModel]] = {}


def register_model(name):
    def register_model_cls(cls):
        if name in MODEL_DICT:
            raise ValueError("Cannot register duplicate trainer ({})".format(name))
        if not issubclass(cls, BaseAutoModel):
            raise ValueError(
                "Trainer ({}: {}) must extend BaseModel".format(name, cls.__name__)
            )
        MODEL_DICT[name] = cls
        return cls

    return register_model_cls


class ModelUniversalRegistry:
    @classmethod
    def get_model(cls, name: str) -> _typing.Type[BaseAutoModel]:
        if type(name) != str:
            raise TypeError(f"Expect model type str, but get {type(name)}.")
        if name not in MODEL_DICT:
            raise KeyError(f"Do not support {name} model in pyg backend")
        return MODEL_DICT.get(name)
